/*
* This file is part of Articulated Scene Model.
*
* Copyright(c) Agnes Swadzba, Leon Ziegler
* http://opensource.cit-ec.de/projects/asm
*
* This file may be licensed under the terms of the
* GNU Lesser General Public License Version 3 (the ``LGPL''),
* or (at your option) any later version.
*
* Software distributed under the License is distributed
* on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
* express or implied. See the LGPL for the specific language
* governing rights and limitations.
*
* You should have received a copy of the LGPL along with this
* program. If not, go to http://www.gnu.org/licenses/lgpl.html
* or write to the Free Software Foundation, Inc.,
* 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
*
* The development of this software was supported by the
* Excellence Cluster EXC 277 Cognitive Interaction Technology.
* The Excellence Cluster EXC 277 is a grant of the Deutsche
* Forschungsgemeinschaft (DFG) in the context of the German
* Excellence Initiative.
*/

#pragma GCC diagnostic ignored "-Wdeprecated-declarations"
#include <asm/ArticulatedSceneComputer.h>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/core/core.hpp>
#include <KGF/kgf.h>
#include <iostream>

using namespace ASM;
using namespace cv;
using namespace std;

int main(int argc, char **argv) {

	int width = 640;
	int height = 480;
	int fps = 30;

	// initialize display
	cv::namedWindow("depth", CV_WINDOW_AUTOSIZE);
	cv::namedWindow("static", CV_WINDOW_AUTOSIZE);
	cv::namedWindow("dynamic", CV_WINDOW_AUTOSIZE);

	// auto select and configure a grabber
	kgf::Grabber::Ptr grabber = kgf::autoSelectGrabber();
	kgf::GrabConfigurationDepth configD(width, height, fps);
	kgf::GrabConfigurationRgb configR(width, height, fps);
	grabber->init(configR, configD);
	grabber->setXyzEnabled(true);
	kgf::Image<kgf::PixelXYZ>::Ptr imgXYZ;
	kgf::Image<kgf::PixelRawDepth>::Ptr imgDepth;

	// initialize Articulated Scene Model
	ArticulatedSceneComputer sceneComputer(width, height);

	cout << "Press ESC to quit" << endl;

	while (true) {

		// get the depthData
		grabber->next();
		imgXYZ = grabber->getLastImageXYZ();
		imgDepth = grabber->getLastImageDepth();

		// convert points to image layers
		XyzImageData::Ptr depthData(new XyzImageData());
		depthData->setWidth(width);
		depthData->setHeight(height);
		for (int y = 0; y < height; y++) {
			for (int x = 0; x < width; x++) {
				kgf::PixelXYZ point = imgXYZ->at(x, y);
				depthData->addX(point.x);
				depthData->addY(point.y);
				depthData->addZ(point.z);
			}
		}

		// calculate scene model
		sceneComputer.work(depthData);

		// get intermediate results
		XyzImageData::Ptr staticData = sceneComputer.getCurrentStatic();
		XyzImageData::Ptr dynamicData = sceneComputer.getCurrentDynamic();

		// convert to opencv Mat
		Mat depthMat(depthData->height(), depthData->width(), CV_32F,
				(float *) depthData->mutableDataZ());
		Mat staticMat(staticData->height(), staticData->width(), CV_32F,
				(float *) staticData->mutableDataZ());
		Mat dynamicMat(dynamicData->height(), dynamicData->width(), CV_32F,
				(float *) dynamicData->mutableDataZ());

		// linearly lower pixel values for visualization
		depthMat /= 4000;
		staticMat /= 4000;
		dynamicMat /= 4000;

		// show images
		cv::imshow("depth", depthMat);
		cv::imshow("static", staticMat);
		cv::imshow("dynamic", dynamicMat);

		char k = cvWaitKey(10);
		if (k == 27) {
			cvDestroyWindow("static");
			cvDestroyWindow("depth");
			cvDestroyWindow("dynamic");
			break;
		}

	}

	return 0;
}
