/*
 * Controller.cpp
 *
 *  Created on: Jul 26, 2012
 *      Author: lziegler
 */

#include "Controller.h"

#include <xcf/Subscriber.hpp>
#include <boost/filesystem.hpp>
#include <sstream>
#include <fstream>
#include <btl/data/navigation/PositionData.h>
#include <xmltio/xmltio.hpp>
#include <KGF/GrabberUtils.h>

#include <pcl/io/pcd_io.h>

using namespace std;
using namespace kgf;
using namespace boost;
using namespace airobots;
using namespace xmltio;
using namespace pcl;

Controller::Controller(Grabber::Ptr grabber, string dataDir) :
		grabber(grabber), dataDir(dataDir) {
}

Controller::~Controller() {
}

void Controller::display(GrabObserver::Ptr listener) {
	grabber->registerObserver(listener);
}

void Controller::unregister(GrabObserver::Ptr listener) {
	grabber->unregisterObserver(listener);
}

void Controller::startRecStatic(string type) {

	string dir = createEmptyDir(type);
	dataRecThread = boost::thread(&Controller::recDataLoop, this, dir);
}

void Controller::stopRecStatic() {
	dataRecThread.interrupt();
	dataRecThread.join();
}

string Controller::createEmptyDir(string name) {

	bool done = false;
	int count = 0;
	filesystem::path complete;
	while (!done) {
		stringstream ss;
		ss << name << count;
		complete = filesystem::path(dataDir) / filesystem::path(ss.str());
		if (exists(complete)) {
			count++;
			continue;
		}
		done = filesystem::create_directory(complete.string());
	}
	return filesystem::absolute(complete).string();
}

void Controller::startRecMoving() {

	string dir = createEmptyDir("moving");
	currentDir = dir;
	dataRecThread = boost::thread(&Controller::recDataLoop, this, dir);
	posRecThread = boost::thread(&Controller::recPositionLoop, this, dir);
}

void Controller::stopRecMoving() {
	dataRecThread.interrupt();
	dataRecThread.join();
	posRecThread.interrupt();
	posRecThread.join();
	currentDir = "";
}

void Controller::setPositionStream(string name) {
	positionStream = name;
}

bool Controller::checkPositionStream() {

	if (positionStream == "") {
		return false;
	}
	try {
		XCF::SubscriberPtr sub = XCF::Subscriber::create(positionStream);
		string content = sub->receiveLast();
		return content != "";

	} catch (XCF::GenericException &e) {
		return false;
	}
	return false;
}

void Controller::startRecPosition() {
	if (currentDir == "") {
		cerr << "Error: data recording must be running in order to "
				"record positions" << endl;
		return;
	}
	if (positionStream == "") {
		cerr << "Error: position stream must be set in order to "
				"record positions" << endl;
		return;
	}

}

void Controller::stopRecPosition() {
}

void Controller::recDataLoop(string dir) {

	ofstream out;
	out.open(string(dir + "/datatimes.dat").c_str());
	out << "# timestamp [msec]     data file" << endl;

	int counter = 0;
	while (true) {
		try {

			stringstream datafile;
			datafile << dir << "/cloud" << setw(5) << setfill('0') << counter++
					<< ".pcd";

			grabber->next();
			PointCloud<PointXYZ>::Ptr cloud = GrabberUtils::toPclCloud(
					grabber->getLastImageXYZ());

			pcl::io::savePCDFileBinary(datafile.str(), *cloud);

			out << GrabberUtils::currentTimeMillis();
			out << "\t";
			out << datafile.str();
			out << endl;
			boost::this_thread::interruption_point();
		} catch (boost::thread_interrupted &e) {
			break;
		}
	}

	out.close();
	cout << "data recording stopped" << endl;
}

void Controller::recPositionLoop(string file) {

	ofstream out;
	out.open(string(file + "/positions.dat").c_str());
	out << "# timestamp [msec]     x [m]     y [m]     theta [rad]" << endl;
	while (true) {
		try {
			try {
				XCF::SubscriberPtr sub = XCF::Subscriber::create(
						positionStream);
				string content = sub->receive();
				Location loc(content, PositionData::getBaseTag());
				PositionData pos;
				PositionData::fromLocation(loc, pos);
				out << pos.getTimestamp().getInserted() << " "
						<< pos.getX(METER) << " " << pos.getY(METER) << " "
						<< pos.getTheta(RADIAN) << endl;

			} catch (XCF::GenericException &e) {
				cerr << "Error while recording positions: " << e.reason << endl;
				break;
			}
			boost::this_thread::interruption_point();
		} catch (boost::thread_interrupted &e) {
			break;
		}
	}
	out.close();
	cout << "position recording stopped";
}
