/*
 * FreenectGrabber.h
 *
 *  Created on: May 30, 2012
 *      Author: lziegler
 */

#ifndef FREENECTGRABBER_H_
#define FREENECTGRABBER_H_
#pragma GCC diagnostic ignored "-Wstrict-aliasing"

#include <libfreenect.h>
#include <libfreenect-registration.h>
#include <string>
#include <stdio.h>
#include <iostream>
#include <log4cxx/logger.h>
#include <cmath>
#include <vector>

#include "Grabber.h"
#include "GrabberUtils.h"

namespace kgf {

/**
 * This is a grabber implementation using the freenect library from the OpenKinect project.
 * http://openkinect.org/
 */
class FreenectGrabber: public Grabber {
private:
	/**
	 * logger instance.
	 */
	log4cxx::LoggerPtr logger;

	/**
	 * libfreenect specific context object.
	 */
	freenect_context *context;
	/**
	 * libfreenect specific device object.
	 */
	freenect_device *device;

	// flags for the grabber's state
	bool imageRgbNew;
	bool imageDepthNew;
	bool calcXyz;
	bool initialized;

	/**
	 * Defines what and how should be grabbed.
	 */
	GrabConfigurationRgb configRgb;
	GrabConfigurationDepth configDepth;

	/**
	 * Thread handling.
	 */
	boost::condition_variable grabRgbCondition;
	boost::condition_variable grabDepthCondition;
	boost::mutex grabRgbMutex;
	boost::mutex grabDepthMutex;

	/**
	 * Image buffers.
	 */
	Image<PixelRGB>::Ptr currentImageRgb;
	Image<PixelRawDepth>::Ptr currentImageDepth;
	Image<PixelXYZ>::Ptr currentImageXyz;

public:
	/**
	 * Default constructor.
	 */
	FreenectGrabber() :
			initialized(false), logger(
					log4cxx::Logger::getLogger("FreenectGrabber")), calcXyz(
					false), stop(false) {
	}

	/**
	 * Destructor.
	 */
	virtual ~FreenectGrabber() {
		LOG4CXX_INFO(logger, "Shutting down FreenectGrabber");
		stop = true;
		if (initialized) {
			freenect_close_device(device);
		}
	}

	/**
	 * Initializes the Grabber.
	 * Creates the interface and starts depth grabbing.
	 */
	void init(GrabConfigurationRgb configRgb,
			GrabConfigurationDepth configDepth) throw (InitializationException) {

		freenect_frame_mode videoMode = extractRgbMode(configRgb);
		freenect_frame_mode depthMode = extractDepthMode(configDepth);

		if (freenect_init(&context, NULL) < 0)
			throw std::runtime_error("Cannot initialize freenect library");
		// We claim both the motor and camera devices, since this class exposes both.
		// It does not support audio, so we do not claim it.
		freenect_select_subdevices(context,
				static_cast<freenect_device_flags>(FREENECT_DEVICE_CAMERA));
		if (pthread_create(&freenectThread, NULL, pthread_callback, this) != 0)
			throw InitializationException("Cannot initialize freenect thread");

		if (freenect_open_device(context, &device, 0))
			throw InitializationException("Can not open Kinect device");

		// add this object as user of the device (required for callbacks)
		freenect_set_user(device, this);

		// initialize required information
		freenect_set_video_mode(device, videoMode);
		freenect_set_depth_mode(device, depthMode);
		freenect_set_depth_callback(device, freenect_depth_callback);
		freenect_set_video_callback(device, freenect_video_callback);

		imageDepthNew = false;
		imageRgbNew = false;

		currentImageRgb = Image<PixelRGB>::Ptr(
				new Image<PixelRGB>(configRgb.resWidth, configRgb.resHeight,
						0));
		currentImageDepth = Image<PixelRawDepth>::Ptr(
				new Image<PixelRawDepth>(configDepth.resWidth,
						configDepth.resHeight, 0));
		currentImageXyz = Image<PixelXYZ>::Ptr(
				new Image<PixelXYZ>(configDepth.resWidth, configDepth.resHeight,
						0));

		// start grabbing
		startDepth();
		startVideo();
		initialized = true;
		LOG4CXX_INFO(logger, "Grabbing started");
	}

	/**
	 * Waits until new (not yet fetched) images are present.
	 */
	void next() throw (InitializationException) {
		if (!initialized) {
			LOG4CXX_WARN(logger, "Grabber was not initialized!");
			throw InitializationException(
					"Grabber Freenect was not initialized!");
			return;
		}
		{
			boost::mutex::scoped_lock lock(grabRgbMutex);
			while (!imageRgbNew) {
				grabRgbCondition.wait(lock);
			}
		}
		{
			boost::mutex::scoped_lock lock(grabDepthMutex);
			while (!imageDepthNew) {
				grabDepthCondition.wait(lock);
			}
		}
	}

	std::string getKey() {
		return "FreenectGrabber";
	}

	Image<PixelRGB>::Ptr getLastImageRGB() {
		boost::mutex::scoped_lock lock(grabRgbMutex);
		imageRgbNew = false;
		return currentImageRgb;
	}

	Image<PixelRawDepth>::Ptr getLastImageDepth() {
		boost::mutex::scoped_lock lock(grabDepthMutex);
		imageDepthNew = false;
		return currentImageDepth;
	}
	Image<PixelXYZ>::Ptr getLastImageXYZ() {
		boost::mutex::scoped_lock lock(grabDepthMutex);
		imageDepthNew = false;
		return currentImageXyz;
	}
	virtual GrabConfigurationRgb getCurrentGrabConfigRgb() {
		return configRgb;
	}
	virtual GrabConfigurationDepth getCurrentGrabConfigDepth() {
		return configDepth;
	}
	virtual std::vector<GrabConfigurationRgb> getAvailableGrabConfigsRgb() {
		std::vector<GrabConfigurationRgb> list;
		list.push_back(GrabConfigurationRgb(640, 480, 30));
		list.push_back(GrabConfigurationRgb(320, 240, 30));
		list.push_back(GrabConfigurationRgb(1280, 1024, 15));
		return list;
	}
	virtual std::vector<GrabConfigurationDepth> getAvailableGrabConfigsDepth() {
		std::vector<GrabConfigurationDepth> list;
		list.push_back(GrabConfigurationDepth(640, 480, 30));
		return list;
	}

	virtual freenect_frame_mode extractRgbMode(GrabConfigurationRgb config)
			throw (InitializationException) {
		if (config.resWidth == 320 && config.resHeight == 240) {
			return freenect_find_video_mode(FREENECT_RESOLUTION_LOW,
					FREENECT_VIDEO_RGB);
		}
		if (config.resWidth == 640 && config.resHeight == 480) {
			return freenect_find_video_mode(FREENECT_RESOLUTION_MEDIUM,
					FREENECT_VIDEO_RGB);
		}
		if (config.resWidth == 1280 && config.resHeight == 1024) {
			return freenect_find_video_mode(FREENECT_RESOLUTION_HIGH,
					FREENECT_VIDEO_RGB);
		}
		throw InitializationException(
				std::string("Desired configuration not supported!\n")
						+ std::string(
								"Available grabbing configurations for rgb:\n")
						+ GrabberUtils::toString(getAvailableGrabConfigsRgb()));
	}
	virtual freenect_frame_mode extractDepthMode(GrabConfiguration config)
			throw (InitializationException) {
		if (config.resWidth == 640 && config.resHeight == 480) {
			return freenect_find_depth_mode(FREENECT_RESOLUTION_MEDIUM,
					FREENECT_DEPTH_MM);
		}
		throw InitializationException(
				std::string("Desired configuration not supported!\n")
						+ std::string(
								"Available grabbing configurations for depth:\n")
						+ GrabberUtils::toString(
								getAvailableGrabConfigsDepth()));
	}

	virtual void setXyzEnabled(bool calc) {
		calcXyz = calc;
	}

	/**
	 * Gets the current tilt angle of kinect.
	 * @return - absolute angle of the kinect [degree]. (0 = horizontal,
	 * 				negative = downwards, positive upwards)
	 */
	double getTiltDegrees() {
		freenect_update_tilt_state(device);
		freenect_raw_tilt_state *state;
		freenect_get_tilt_state(device);
		return freenect_get_tilt_degs(state);
	}

	/**
	 * Tells the client if this implementation can provide the camera's angle.
	 * @return true if implementation can provide angle.
	 */
	bool canGetTilt() {
		return true;
	}

	/**
	 * Enables or disables grabbing from camera.
	 * @param enabled
	 */
	inline void setEnabled(bool enabled) {
		try {
			if (enabled) {
				startVideo();
				startDepth();
			} else {
				stopVideo();
				stopDepth();
			}
		} catch (InitializationException &e) {
			LOG4CXX_ERROR(logger,
					"Something went wrong enabling or disabling the grabbing");
		}
	}

	void callbackRgb(uint8_t *rgb, uint32_t timestamp) {

		int w = configRgb.resWidth;
		int h = configRgb.resHeight;
		int totalSize = w * h;

		grabRgbMutex.lock();

		// both (uint8_t,uint8_t,uint8_t) and PixelRGB are represented
		// similarly in memory
		// THIS POINTER MUST NOT BE DELETED!!
		PixelRGB *pixels = (PixelRGB*) rgb;

		// create image structure. This constructor copies the pixels internally.
		currentImageRgb = Image<PixelRGB>::Ptr(
				new Image<PixelRGB>(w, h, pixels, timestamp));

		notifyObserversRgb(currentImageRgb);
		imageRgbNew = true;
		grabRgbMutex.unlock();

		// notify the next() function that new data is present
		grabRgbCondition.notify_all();
	}
	void callbackDepth(uint16_t *depth, uint32_t timestamp) {

		int w = configDepth.resWidth;
		int h = configDepth.resHeight;
		int totalSize = w * h;

		grabDepthMutex.lock();

		// both uint16_t and PixelRGB are represented
		// similarly in memory
		// THIS POINTER MUST NOT BE DELETED!!
		PixelRawDepth *pixels = (PixelRawDepth*) depth;

		// create image structure. This constructor copies the pixels internally.
		currentImageDepth = Image<PixelRawDepth>::Ptr(
				new Image<PixelRawDepth>(w, h, pixels, timestamp));

		imageDepthNew = true;

		// calculate point cloud, if desired
		if (calcXyz) {
			Image<PixelXYZ>::Ptr cloud(new Image<PixelXYZ>(w, h, timestamp));
			for (int v = 0, n = 0; v < h; v++) {
				for (int u = 0; u < w; u++, n++) {
					// calculate points from depth and camera calibration.
					uint16_t d = depth[n];
					PixelXYZ point;
					if (d == 0) {
						point.x = 0;
						point.y = 0;
						point.z = 0;
					} else {
						double wx, wy;
						freenect_camera_to_world(device, u, v, d, &wx, &wy);
						point.x = wx;
						point.y = wy;
						point.z = d;
					}
					cloud->at(u, v) = point;
				}
			}
			currentImageXyz.swap(cloud);
		}
		grabDepthMutex.unlock();

		notifyObserversDepth(currentImageDepth);
		// notify the next() function that new data is present
		grabDepthCondition.notify_all();
	}

private:

	volatile bool stop;
	pthread_t freenectThread;

	static void freenect_depth_callback(freenect_device *dev, void *depth,
			uint32_t timestamp) {
		FreenectGrabber* grabber =
				static_cast<FreenectGrabber*>(freenect_get_user(dev));
		grabber->callbackDepth(static_cast<uint16_t*>(depth), timestamp);
	}
	static void freenect_video_callback(freenect_device *dev, void *video,
			uint32_t timestamp) {
		FreenectGrabber* grabber =
				static_cast<FreenectGrabber*>(freenect_get_user(dev));
		grabber->callbackRgb(static_cast<uint8_t*>(video), timestamp);
	}
	static void *pthread_callback(void *user_data) {
		FreenectGrabber* grabber = static_cast<FreenectGrabber*>(user_data);
		grabber->freenectMainLoop();
		return NULL;
	}
	void startVideo() throw (InitializationException) {
		if (freenect_start_video(device) < 0)
			throw InitializationException("Cannot start RGB callback");
	}
	void stopVideo() throw (InitializationException) {
		if (freenect_stop_video(device) < 0)
			throw InitializationException("Cannot stop RGB callback");
	}
	void startDepth() throw (InitializationException) {
		if (freenect_start_depth(device) < 0)
			throw InitializationException("Cannot start depth callback");
	}
	void stopDepth() throw (InitializationException) {
		if (freenect_stop_depth(device) < 0)
			throw InitializationException("Cannot stop depth callback");
	}
	void freenectMainLoop() {
		while (!stop) {
			if (freenect_process_events(context) < 0)
				throw std::runtime_error("Cannot process freenect events");
		}
	}
	virtual void registerObserver(GrabObserver::Ptr observer) {
		registerObserverImpl(observer);
	}
	virtual void unregisterObserver(GrabObserver::Ptr observer) {
		unregisterObserverImpl(observer);
	}
}
;

}
#endif /* FREENECTGRABBER_H_ */
