/*
 * Grabber.h
 *
 *  Created on: Jul 16, 2012
 *      Author: lziegler
 */

#ifndef GRABBER_H_
#define GRABBER_H_

#include "GrabberTypes.h"
#include <vector>
#include <exception>
#include <boost/thread.hpp>

namespace kgf {

/**
 * Interface for clients using the callback procedure of the grabber. All
 * clients that need to be notified of new frames should implement this
 * interface and register themselves as observers to the grabber.
 */
class GrabObserver {
public:
	/**
	 * Notification function for new color images.
	 * @param A pointer at the actual color image data.
	 */
	virtual void processImageRGB(Image<PixelRGB>::Ptr) = 0;
	/**
	 * Notification function for new depth images.
	 * @param A pointer at the actual depth image data.
	 */
	virtual void processImageDepth(Image<PixelRawDepth>::Ptr) = 0;
	/**
	 * Notification function for new xyz images.
	 * @param A pointer at the actual xyz image data.
	 */
	virtual void processImageXYZ(Image<PixelXYZ>::Ptr) = 0;

	typedef boost::shared_ptr<GrabObserver> Ptr;
	virtual ~GrabObserver() {
	}
};

/**
 * A container class for holding a set of parameters describing the
 * configuration of the grabber.
 */
class GrabConfiguration {
public:
	int resWidth;
	int resHeight;
	int fps;
	GrabConfiguration() :
			resWidth(640), resHeight(480), fps(30) {
	}
	GrabConfiguration(int resWidth, int resHeight, int fps) :
			resWidth(resWidth), resHeight(resHeight), fps(fps) {
	}
};
class GrabConfigurationRgb: public GrabConfiguration {
public:
	GrabConfigurationRgb() :
			GrabConfiguration() {
	}
	GrabConfigurationRgb(int resWidth, int resHeight, int fps) :
			GrabConfiguration(resWidth, resHeight, fps) {
	}
};
class GrabConfigurationDepth: public GrabConfiguration {
public:
	GrabConfigurationDepth() :
			GrabConfiguration() {
	}
	GrabConfigurationDepth(int resWidth, int resHeight, int fps) :
			GrabConfiguration(resWidth, resHeight, fps) {
	}
};

/**
 * Exception for reporting errors while creating the grabber.
 */
class GrabberException: std::exception {
public:
	GrabberException() {
	}
	GrabberException(std::string msg) {
		this->msg = msg;
	}
	const char* what() {
		return msg.c_str();
	}
	virtual ~GrabberException() throw () {
	}
private:
	std::string msg;
};

/**
 * Exception for reporting errors while initializing the grabber.
 */
class InitializationException: std::exception {
public:
	InitializationException() {
	}
	InitializationException(std::string msg) {
		this->msg = msg;
	}
	const char* what() {
		return msg.c_str();
	}
	virtual ~InitializationException() throw () {
	}
private:
	std::string msg;
};

/**
 * This is the general interface for any grabber implementation in kgf.
 */
class Grabber {
public:
	typedef boost::shared_ptr<Grabber> Ptr;
	Grabber() {
	}
	virtual ~Grabber() {
	}

	////////////////////////// General methods /////////////////////////

	/**
	 * Initialize and configure the grabber.
	 * @param configRgb Abstract description of the desired grabbing properties.
	 * @param configDepth Abstract description of the desired grabbing properties.
	 */
	virtual void init(GrabConfigurationRgb configRgb,
			GrabConfigurationDepth configDepth) throw (InitializationException) = 0;

	/**
	 * Enable/disable grabbing.
	 * @param enabled If true: enable. Otherwise: disable
	 */
	virtual void setEnabled(bool enabled) = 0;

	/**
	 * Getter for a key string representing the specific grabber implementation.
	 * @return A key string representing the specific grabber implementation.
	 */
	virtual std::string getKey() = 0;

	/**
	 * Getter for the current configuration of the grabber.
	 * @return Abstract description of the desired grabbing properties.
	 */
	virtual GrabConfigurationRgb getCurrentGrabConfigRgb() = 0;
	/**
	 * Getter for the current configuration of the grabber.
	 * @return Abstract description of the desired grabbing properties.
	 */
	virtual GrabConfigurationDepth getCurrentGrabConfigDepth() = 0;

	/**
	 * Getter for a list of all configuration that the specific implementation in combination with
	 * the connected hardware can process.
	 * @return A list of GrabConfiguration objects.
	 */
	virtual std::vector<GrabConfigurationRgb> getAvailableGrabConfigsRgb() = 0;
	/**
	 * Getter for a list of all configuration that the specific implementation in combination with
	 * the connected hardware can process.
	 * @return A list of GrabConfiguration objects.
	 */
	virtual std::vector<GrabConfigurationDepth> getAvailableGrabConfigsDepth() = 0;

	/**
	 * Enable/disable the calculation of the xyz-image. Attention if enabled, this costs
	 * significant processing power.
	 * @param enabled If true: enable. Otherwise: disable
	 */
	virtual void setXyzEnabled(bool enabled) = 0;

	virtual bool canGetTilt() = 0;
	virtual double getTiltDegrees() = 0;

	////////////////////////// Synchronous interface /////////////////////////

	/**
	 * Grab next frame. This is part of the synchronous interface. Call this everytime you want to
	 * receive a new frame.
	 */
	virtual void next() throw (InitializationException) = 0;

	/**
	 * This is a getter for the last received color image. In order to receive the latest frame,
	 * next() will have to be called first.
	 *
	 * @return A pointer to the image.
	 */
	virtual Image<PixelRGB>::Ptr getLastImageRGB() = 0;

	/**
	 * This is a getter for the last received depth image. In order to receive the latest frame,
	 * next() will have to be called first. Values are in millimeter [mm].
	 *
	 * @return A pointer to the image. Pixel values are in millimeter [mm].
	 */
	virtual Image<PixelRawDepth>::Ptr getLastImageDepth() = 0;

	/**
	 * This is a getter for the point cloud image of the last received depth image. In order to
	 * receive the latest frame, next() will have to be called first.
	 *
	 * @return A pointer to the image. Values are in millimeter [mm].
	 */
	virtual Image<PixelXYZ>::Ptr getLastImageXYZ() = 0;

	////////////////////////// Asynchronous interface /////////////////////////

	virtual void registerObserver(GrabObserver::Ptr observer) = 0;
	virtual void unregisterObserver(GrabObserver::Ptr observer) = 0;

protected:
	/**
	 * Observers for new data.
	 */
	std::vector<GrabObserver::Ptr> observers;

	void notifyObserversRgb(Image<PixelRGB>::Ptr img) {
		for (std::vector<GrabObserver::Ptr>::iterator o = observers.begin();
				o != observers.end(); ++o) {
			GrabObserver::Ptr obs = *o;
			boost::thread(&GrabObserver::processImageRGB, obs.get(), img);
		}
	}
	void notifyObserversDepth(Image<PixelRawDepth>::Ptr img) {
		for (std::vector<GrabObserver::Ptr>::iterator o = observers.begin();
				o != observers.end(); ++o) {
			GrabObserver::Ptr obs = *o;
			boost::thread(&GrabObserver::processImageDepth, obs.get(), img);
		}
	}
	void notifyObserversXyz(Image<PixelXYZ>::Ptr img) {
		for (std::vector<GrabObserver::Ptr>::iterator o = observers.begin();
				o != observers.end(); ++o) {
			GrabObserver::Ptr obs = *o;
			boost::thread(&GrabObserver::processImageXYZ, obs.get(), img);
		}
	}
	void registerObserverImpl(GrabObserver::Ptr observer) {
		observers.push_back(observer);
	}
	void unregisterObserverImpl(GrabObserver::Ptr observer) {
		std::vector<GrabObserver::Ptr>::iterator item = std::find(
				observers.begin(), observers.end(), observer);
		if (item != observers.end()) {
			observers.erase(item);
		}
	}
};
}
#endif /* GRABBER_H_ */
