/*
 * Grabber.h
 *
 *  Created on: Jul 16, 2012
 *      Author: lziegler
 */

#ifndef GRABBER_TYPES_H_
#define GRABBER_TYPES_H_

#include <boost/shared_ptr.hpp>
#include <vector>
#include <string.h>

namespace kgf {

/**
 * Defines a raw measured depth value.
 */
typedef unsigned short PixelRawDepth;

/**
 * Defines a measured depth value in [mm].
 */
typedef float PixelRealDepth;

/**
 * Defines a RGB pixel value.
 */
typedef struct PixelRGB {
	unsigned char nRed;
	unsigned char nGreen;
	unsigned char nBlue;
} PixelRGB;

/**
 * Defines a XYZ pixel value in [mm].
 */
typedef struct PixelXYZ {
	float x;
	float y;
	float z;
} PixelXYZ;

/**
 * Defines an grabbed image
 */
template<typename PixelT>
class Image {
private:
	unsigned long timestamp;
	unsigned int width;
	unsigned int height;
	PixelT* data;
public:

	Image() :
			timestamp(0), width(0), height(0), data(0) {
	}
	Image(unsigned int width, unsigned int height, unsigned long timestamp) :
			timestamp(timestamp), width(width), height(height), data(
					new PixelT[width * height]) {
	}
	Image(unsigned int width, unsigned int height, const PixelT* data_in,
			unsigned long timestamp, bool nocopy = false) :
			timestamp(timestamp), width(width), height(height) {

		if (nocopy) {
			this->data = const_cast<PixelT*>(data_in);
		} else {
			this->data = new PixelT[width * height];
			memcpy(this->data, data_in, width * height * sizeof(PixelT));
		}
	}

	virtual ~Image() {
		delete data;
	}

	inline const unsigned long getTimestamp() const {
		return timestamp;
	}

	inline const unsigned int getWidth() const {
		return width;
	}

	inline const unsigned int getHeight() const {
		return height;
	}

	inline const PixelT&
	at(int column, int row) const {
		return (data[row * this->width + column]);
	}

	inline PixelT&
	at(int column, int row) {
		return (data[row * this->width + column]);
	}

	inline const PixelT&
	operator ()(size_t column, size_t row) const {
		return (data[row * this->width + column]);
	}

	inline PixelT&
	operator ()(size_t column, size_t row) {
		return (data[row * this->width + column]);
	}

	typedef PixelT PointType; // Make the template class available from the outside
	typedef boost::shared_ptr<Image<PixelT> > Ptr;
	typedef boost::shared_ptr<const Image<PixelT> > ConstPtr;

	// iterators
	typedef typename std::vector<PixelT>::iterator iterator;
	typedef typename std::vector<PixelT>::const_iterator const_iterator;
	inline iterator begin() {
		return (data);
	}
	inline iterator end() {
		return (data + width * height * sizeof(PixelT));
	}
	inline const_iterator begin() const {
		return (data);
	}
	inline const_iterator end() const {
		return (data + width * height * sizeof(PixelT));
	}

	//capacity
	inline size_t size() const {
		return (width * height);
	}
	inline bool empty() const {
		return width == 0 && height == 0;
	}

	//element access
	inline const PixelT& operator[](size_t n) const {
		return (data[n]);
	}
	inline PixelT& operator[](size_t n) {
		return (data[n]);
	}
	inline const PixelT& at(size_t n) const {
		return (data[n]);
	}
	inline PixelT& at(size_t n) {
		return (data[n]);
	}
	inline const PixelT& front() const {
		return (data[0]);
	}
	inline PixelT& front() {
		return (data[0]);
	}
	inline const PixelT& back() const {
		return (data[width * height]);
	}
	inline PixelT& back() {
		return (data[width * height]);
	}

	inline PixelT* data_ptr() const {
		return data;
	}
};
}
#endif /* GRABBER_TYPES_H_ */
