/*
 * OpenNIGrabber.h
 *
 *  Created on: Feb 1, 2011
 *      Author: lziegler
 */

#ifndef OPENNIGRABBER_H_
#define OPENNIGRABBER_H_

#pragma GCC diagnostic ignored "-Wreorder"
#pragma GCC diagnostic ignored "-Wstrict-aliasing"
#pragma GCC diagnostic ignored "-Wunused-variable"
#pragma GCC diagnostic ignored "-Wreturn-type"
#include <XnOpenNI.h>
#include <XnCodecIDs.h>
#include <XnCppWrapper.h>
#include <vector>
#include <boost/thread.hpp>
#include <boost/thread/condition_variable.hpp>
#include <log4cxx/logger.h>

#include "Grabber.h"

namespace kgf {

class OpenNIGrabber: public Grabber {
protected:
	/**
	 * OpenNI objects.
	 */
	xn::Context context;
	xn::DepthGenerator depthGen;
	xn::ImageGenerator imageGen;

	/**
	 * Defines what and how should be grabbed.
	 */
	GrabConfigurationRgb configRgb;
	GrabConfigurationDepth configDepth;

	/**
	 * Thread handling.
	 */
	boost::thread grabAllThread;
	boost::condition_variable grabAllCondition;
	boost::mutex grabRgbMutex;
	boost::mutex grabDepthMutex;
	boost::mutex waitAllMutex;

	/**
	 * Image buffers.
	 */
	Image<PixelRGB>::Ptr currentImageRgb;
	Image<PixelRawDepth>::Ptr currentImageDepth;
	Image<PixelXYZ>::Ptr currentImageXyz;

	// flags for the grabber's state
	bool imageRgbNew;
	bool imageDepthNew;
	bool initialized;
	bool calcXyz;

public:
	/**
	 * Constructs a new grabber instance.
	 */
	OpenNIGrabber() :
			Grabber(), logger(log4cxx::Logger::getLogger("OpenNIGrabber")), initialized(
					false), imageRgbNew(false), imageDepthNew(false), calcXyz(
					false) {
		LOG4CXX_INFO(logger, "Creating OpenNIGrabber");
	}

	/**
	 * Waits until new (not yet fetched) images are present.
	 */
	virtual void next() throw (InitializationException) {
		if (!initialized) {
			LOG4CXX_WARN(logger, "Grabber was not initialized!");
			throw InitializationException("Grabber Freenect was not initialized!");
			return;
		}
		if (!imageDepthNew || !imageRgbNew) {
			boost::mutex::scoped_lock lock(waitAllMutex);
			grabAllCondition.wait(lock);
		}
	}

	virtual std::string getKey() {
		return "OpenNIGrabber";
	}

	virtual Image<PixelRGB>::Ptr getLastImageRGB() {
		boost::mutex::scoped_lock lock(grabRgbMutex);
		imageRgbNew = false;
		return currentImageRgb;
	}
	virtual Image<PixelRawDepth>::Ptr getLastImageDepth() {
		boost::mutex::scoped_lock lock(grabDepthMutex);
		imageDepthNew = false;
		return currentImageDepth;
	}
	virtual Image<PixelXYZ>::Ptr getLastImageXYZ() {
		boost::mutex::scoped_lock lock(grabDepthMutex);
		imageDepthNew = false;
		return currentImageXyz;
	}
	virtual GrabConfigurationRgb getCurrentGrabConfigRgb() {
		return configRgb;
	}
	virtual GrabConfigurationDepth getCurrentGrabConfigDepth() {
		return configDepth;
	}
	virtual std::vector<GrabConfigurationRgb> getAvailableGrabConfigsRgb() {
		std::vector<GrabConfigurationRgb> list;
		list.push_back(GrabConfigurationRgb(1280, 1024, 15));
		list.push_back(GrabConfigurationRgb(640, 480, 30));
		list.push_back(GrabConfigurationRgb(640, 480, 25));
		list.push_back(GrabConfigurationRgb(320, 240, 25));
		list.push_back(GrabConfigurationRgb(320, 240, 30));
		list.push_back(GrabConfigurationRgb(320, 240, 60));
		list.push_back(GrabConfigurationRgb(160, 120, 25));
		list.push_back(GrabConfigurationRgb(160, 120, 30));
		list.push_back(GrabConfigurationRgb(160, 120, 60));
		return list;
	}
	virtual std::vector<GrabConfigurationDepth> getAvailableGrabConfigsDepth() {
		std::vector<GrabConfigurationDepth> list;
		list.push_back(GrabConfigurationDepth(1280, 1024, 15));
		list.push_back(GrabConfigurationDepth(640, 480, 30));
		list.push_back(GrabConfigurationDepth(640, 480, 25));
		list.push_back(GrabConfigurationDepth(320, 240, 25));
		list.push_back(GrabConfigurationDepth(320, 240, 30));
		list.push_back(GrabConfigurationDepth(320, 240, 60));
		list.push_back(GrabConfigurationDepth(160, 120, 25));
		list.push_back(GrabConfigurationDepth(160, 120, 30));
		list.push_back(GrabConfigurationDepth(160, 120, 60));
		return list;
	}
	virtual void setXyzEnabled(bool calc) {
		calcXyz = calc;
	}

	virtual void init(GrabConfigurationRgb configRgb,
			GrabConfigurationDepth configDepth) throw (InitializationException) {
		LOG4CXX_INFO(logger, "Initializing OpenNILogger");
		this->configRgb = configRgb;
		this->configDepth = configDepth;

		// init OpenNI
		XnStatus nRetVal = XN_STATUS_OK;

		// Initialize context object
		nRetVal = context.Init();
		checkStatus(nRetVal, "init");

		// Create generator nodes
		nRetVal = depthGen.Create(context);
		checkStatus(nRetVal, "create depth generator");

		// Set output mode
		XnMapOutputMode outputModeRgb;
		outputModeRgb.nXRes = configRgb.resWidth;
		outputModeRgb.nYRes = configRgb.resHeight;
		outputModeRgb.nFPS = configRgb.fps;
		XnMapOutputMode outputModeDepth;
		outputModeDepth.nXRes = configDepth.resWidth;
		outputModeDepth.nYRes = configDepth.resHeight;
		outputModeDepth.nFPS = configDepth.fps;

		nRetVal = depthGen.SetMapOutputMode(outputModeDepth);
		checkStatus(nRetVal);

		// Create generator nodes
		nRetVal = imageGen.Create(context);
		checkStatus(nRetVal, "create image generator");

		nRetVal = imageGen.SetMapOutputMode(outputModeRgb);
		nRetVal = imageGen.SetPixelFormat(XN_PIXEL_FORMAT_RGB24);
		checkStatus(nRetVal);

		imageDepthNew = false;
		imageRgbNew = false;

		// Make it start generating data
		nRetVal = context.StartGeneratingAll();
		checkStatus(nRetVal, "start generating");

		initialized = true;

		// start the threads, that actually grab from the generator nodes
		grabAllThread = boost::thread(&OpenNIGrabber::grabAllLoop, this);
		LOG4CXX_INFO(logger, "grabbing started");
	}

	virtual void grabAllLoop() {
		while (true) {
			try {
				// wait for new data
				context.WaitAndUpdateAll();
				grabDepth();
				grabRgb();
				boost::this_thread::interruption_point();
				// notify the next() function that new data is present
				grabAllCondition.notify_all();
			} catch (boost::thread_interrupted &e) {
				break;
			}
		}
	}

	virtual void grabRgb() {
		xn::ImageMetaData meta;
		int w = configRgb.resWidth;
		int h = configRgb.resHeight;

		// get image as RGB. THIS POINTER MUST NOT BE DELETED!!
		const XnRGB24Pixel* map = imageGen.GetRGB24ImageMap();

		// get timestamp
		imageGen.GetMetaData(meta);
		long time = meta.Timestamp();

		// both XnRGB24Pixel and PixelRGB are represented
		// similarly in memory
		// THIS POINTER MUST NOT BE DELETED!!
		const PixelRGB* pixels = (const PixelRGB*) map;

		boost::mutex::scoped_lock lock(grabRgbMutex);
		// create image structure. This constructor copies the pixels internally.
		Image<PixelRGB>::Ptr imageRgb(new Image<PixelRGB>(w, h, pixels, time));
		currentImageRgb = imageRgb;
		imageRgbNew = true;
		// notify the observers
		notifyObserversRgb(imageRgb);

	}

	virtual void grabDepth() {
		xn::DepthMetaData meta;
		int w = configDepth.resWidth;
		int h = configDepth.resHeight;
		int total = w * h;

		const XnDepthPixel* map = depthGen.GetDepthMap();

		// get timestamp
		depthGen.GetMetaData(meta);
		long time = meta.Timestamp();

		// both XnDepthPixel and PixelRawDepth are represented
		// similarly in memory
		// THIS POINTER MUST NOT BE DELETED!! (OpenNI is owner)
		PixelRawDepth* pixels = (PixelRawDepth*) map;

		boost::mutex::scoped_lock lock(grabDepthMutex);
		// create image structure. This constructor copies the pixels internally.
		Image<PixelRawDepth>::Ptr imageDepth(
				new Image<PixelRawDepth>(w, h, pixels, time));

		currentImageDepth = imageDepth;
		notifyObserversDepth(imageDepth);

		if (calcXyz) {
			// create xyz image

			// temporary data structure in scope
			XnPoint3D projective[total];
			for (int u = 0; u < w; u++) {
				for (int v = 0; v < h; v++) {
					int i(v * w + u);
					XnPoint3D p;
					p.X = (float) u;
					p.Y = (float) v;
					p.Z = (float) currentImageDepth->at(u, v);
					projective[i] = p;
				}
			}

			// long-lasting data structure on heap, must be deleted.
			XnPoint3D* realworld = new XnPoint3D[total];

			// convert
			depthGen.ConvertProjectiveToRealWorld(total, projective, realworld);

			// both XnPoint3D and PixelXYZ are represented
			// similarly in memory
			PixelXYZ* points = (PixelXYZ*) realworld;

			// create image structure. This constructor does NOT copy the points, takes
			// ownership and will delete the data.
			currentImageXyz = Image<PixelXYZ>::Ptr(
					new Image<PixelXYZ>(w, h, points, time, true));
			notifyObserversXyz(currentImageXyz);
		}
		imageDepthNew = true;
	}

	virtual inline bool canGetTilt() {
		return false;
	}

	virtual inline double getTiltDegrees() {
		return 0;
	}

	virtual void setEnabled(bool enabled) {
		XnStatus nRetVal = XN_STATUS_OK;
		if (enabled) {
			if (!initialized) {
				// Make it start generating data
				nRetVal = context.StartGeneratingAll();
				checkStatus(nRetVal, "start generating");
				initialized = true;
			}
		} else {
			if (initialized) {
				context.StopGeneratingAll();
				checkStatus(nRetVal, "stop generating");
				initialized = false;
			}
		}
	}

	virtual ~OpenNIGrabber() {
		LOG4CXX_INFO(logger, "Shutting down OpenNI");

		grabAllThread.interrupt();
		boost::posix_time::ptime deadline =
				boost::posix_time::microsec_clock::local_time()
						+ boost::posix_time::seconds(1);
		grabAllThread.timed_join(deadline);

		setEnabled(false);
		context.Release();
	}

protected:
	log4cxx::LoggerPtr logger; /**<logger instance.*/

	void printError(uint statusCode, std::string action = "") {
		std::string error = xnGetStatusString(statusCode);
		throw InitializationException(error);
		LOG4CXX_ERROR(logger, "OpenNI Error (" << action <<"): " << error);
		if (statusCode == XN_STATUS_NO_NODE_PRESENT) {
			LOG4CXX_ERROR(logger, "Hardware device not connected?");
			LOG4CXX_ERROR(logger, "OPEN_NI_INSTALL_PATH set correctly?");
		}
		if (statusCode == XN_STATUS_OS_NETWORK_TIMEOUT) {
			LOG4CXX_ERROR(logger, "Try killing process XnSensorServer");
		}
	}
	void checkStatus(uint statusCode, std::string action = "") {
		if (statusCode != XN_STATUS_OK) {
			printError(statusCode, action);
		}
	}
	virtual void registerObserver(GrabObserver::Ptr observer) {
		registerObserverImpl(observer);
	}
	virtual void unregisterObserver(GrabObserver::Ptr observer) {
		unregisterObserverImpl(observer);
	}
};
}
#endif /* OPENNIGRABBER_H_ */
