/*
* This file is part of Articulated Scene Model.
*
* Copyright(c) Agnes Swadzba, Leon Ziegler
* http://opensource.cit-ec.de/projects/asm
*
* This file may be licensed under the terms of the
* GNU Lesser General Public License Version 3 (the ``LGPL''),
* or (at your option) any later version.
*
* Software distributed under the License is distributed
* on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
* express or implied. See the LGPL for the specific language
* governing rights and limitations.
*
* You should have received a copy of the LGPL along with this
* program. If not, go to http://www.gnu.org/licenses/lgpl.html
* or write to the Free Software Foundation, Inc.,
* 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
*
* The development of this software was supported by the
* Excellence Cluster EXC 277 Cognitive Interaction Technology.
* The Excellence Cluster EXC 277 is a grant of the Deutsche
* Forschungsgemeinschaft (DFG) in the context of the German
* Excellence Initiative.
*/

#include "ArticulatedSceneComputer.h"
#include "XyzImageData.h"

#include <boost/thread/mutex.hpp>

using namespace std;
using namespace boost;
using namespace cv;

namespace ASM {

ArticulatedSceneComputer::ArticulatedSceneComputer(int width, int height) :
		scene(new ArticulatedScene(width, height)), width(width), height(height) {
}

ArticulatedSceneComputer::~ArticulatedSceneComputer() {
}

void ArticulatedSceneComputer::reset() {

	boost::mutex::scoped_lock lock(workMutex);

	scene->staticX.setTo(cv::Scalar(0));
	scene->staticY.setTo(cv::Scalar(0));
	scene->staticZ.setTo(cv::Scalar(0));

	scene->weights.setTo(cv::Scalar(0));
}

ArticulatedScene::Ptr ArticulatedSceneComputer::getCurrentSceneCopy() const {
	mutex::scoped_lock lock(workMutex);
	ArticulatedScene::Ptr copy(new ArticulatedScene(scene));
	return copy;
}

XyzImageData::Ptr ArticulatedSceneComputer::getCurrentStatic() const {
	mutex::scoped_lock lock(resultMutex);
	return currentStatic;
}

XyzImageData::Ptr ArticulatedSceneComputer::getCurrentDynamic() const {
	mutex::scoped_lock lock(resultMutex);
	return currentDynamic;
}

void ArticulatedSceneComputer::initialize(ArticulatedScene::Ptr scene) {

	mutex::scoped_lock lock(workMutex);

	this->scene = scene;
}

void ArticulatedSceneComputer::initialize(XyzImageData::Ptr img) {

	cv::Mat xMat(img->height(), img->width(), CvMatType, img->mutableDataX());
	cv::Mat yMat(img->height(), img->width(), CvMatType, img->mutableDataY());
	cv::Mat zMat(img->height(), img->width(), CvMatType, img->mutableDataZ());
	initialize(xMat, yMat, zMat);
}

void ArticulatedSceneComputer::initialize(Mat xMat, Mat yMat, Mat zMat) {

	mutex::scoped_lock lock(workMutex);

	// compute valid depth measurements (z != 0)
	cv::Mat validMat;
	cv::compare(zMat, 0.0, validMat, cv::CMP_NE);

	xMat.copyTo(scene->staticX);
	yMat.copyTo(scene->staticY);
	zMat.copyTo(scene->staticZ);

	scene->weights.setTo(cv::Scalar(1), validMat);
}

void ArticulatedSceneComputer::work(const XyzImageData::Ptr img) {

	cv::Mat xMat(img->height(), img->width(), CvMatType, img->mutableDataX());
	cv::Mat yMat(img->height(), img->width(), CvMatType, img->mutableDataY());
	cv::Mat zMat(img->height(), img->width(), CvMatType, img->mutableDataZ());
	work(xMat, yMat, zMat);
}

void ArticulatedSceneComputer::work(const Mat &xMat, const Mat &yMat,
		const Mat &zMat) {

	float diff, depth, thres;

	float cx, cy, cz;
	float sx, sy, sz;

	{
		mutex::scoped_lock lock(workMutex);

		scene->dynamicX.setTo(cv::Scalar(0));
		scene->dynamicY.setTo(cv::Scalar(0));
		scene->dynamicZ.setTo(cv::Scalar(0));

		for (int y = 0; y < height; y++) {

			for (int x = 0; x < width; x++) {

				// current depth measurement is invalid
				if (zMat.at<float>(y, x) == 0.0) {
					continue;
				}

				cx = xMat.at<float>(y, x);
				cy = yMat.at<float>(y, x);
				cz = zMat.at<float>(y, x);

				sx = scene->staticX.at<float>(y, x);
				sy = scene->staticY.at<float>(y, x);
				sz = scene->staticZ.at<float>(y, x);

				// compute Euclidean between current point and the corresponding point in the static model
				depth = sqrt(cx * cx + cy * cy + cz * cz);
				diff = depth - sqrt(sx * sx + sy * sy + sz * sz);
				thres = 0.1 * depth;

				if (scene->weights.at<float>(y, x) == 0) { // initialize static model with current measurement
					scene->staticX.at<float>(y, x) = xMat.at<float>(y, x);
					scene->staticY.at<float>(y, x) = yMat.at<float>(y, x);
					scene->staticZ.at<float>(y, x) = zMat.at<float>(y, x);
					scene->weights.at<float>(y, x) = 1.0;
				} else if (abs(diff) < thres) { // current measurement is in range of known static measurement
					scene->staticX.at<float>(y, x) += 1.0
							/ scene->weights.at<float>(y, x)
							* (xMat.at<float>(y, x)
									- scene->staticX.at<float>(y, x));
					scene->staticY.at<float>(y, x) += 1.0
							/ scene->weights.at<float>(y, x)
							* (yMat.at<float>(y, x)
									- scene->staticY.at<float>(y, x));
					scene->staticZ.at<float>(y, x) += 1.0
							/ scene->weights.at<float>(y, x)
							* (zMat.at<float>(y, x)
									- scene->staticZ.at<float>(y, x));
					scene->weights.at<float>(y, x) += 1.0;
				} else if (diff > 0.0) { // current measurement is further away than known static value
					scene->staticX.at<float>(y, x) = xMat.at<float>(y, x);
					scene->staticY.at<float>(y, x) = yMat.at<float>(y, x);
					scene->staticZ.at<float>(y, x) = zMat.at<float>(y, x);
					scene->weights.at<float>(y, x) = 1.0;
				} else { // current measurement is in front of known static value
					scene->dynamicX.at<float>(y, x) = xMat.at<float>(y, x);
					scene->dynamicY.at<float>(y, x) = yMat.at<float>(y, x);
					scene->dynamicZ.at<float>(y, x) = zMat.at<float>(y, x);
				}
			}
		}

		// add updated static and dynamic value to intermediate results structure
		XyzImageData::Ptr staticImage(new XyzImageData(width, height));
		XyzImageData::Ptr dynamicImage(new XyzImageData(width, height));
		memcpy(staticImage->mutableDataX(),scene->staticX.data, sizeof(float) * width* height);
		memcpy(staticImage->mutableDataY(),scene->staticY.data, sizeof(float) * width* height);
		memcpy(staticImage->mutableDataZ(),scene->staticZ.data, sizeof(float) * width* height);
		memcpy(dynamicImage->mutableDataX(),scene->dynamicX.data, sizeof(float) * width* height);
		memcpy(dynamicImage->mutableDataY(),scene->dynamicY.data, sizeof(float) * width* height);
		memcpy(dynamicImage->mutableDataZ(),scene->dynamicZ.data, sizeof(float) * width* height);

		mutex::scoped_lock resultlock(resultMutex);
		currentDynamic = dynamicImage;
		currentStatic = staticImage;
	}
}
}
