/*
* This file is part of Articulated Scene Model.
*
* Copyright(c) Agnes Swadzba, Leon Ziegler
* http://opensource.cit-ec.de/projects/asm
*
* This file may be licensed under the terms of the
* GNU Lesser General Public License Version 3 (the ``LGPL''),
* or (at your option) any later version.
*
* Software distributed under the License is distributed
* on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
* express or implied. See the LGPL for the specific language
* governing rights and limitations.
*
* You should have received a copy of the LGPL along with this
* program. If not, go to http://www.gnu.org/licenses/lgpl.html
* or write to the Free Software Foundation, Inc.,
* 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
*
* The development of this software was supported by the
* Excellence Cluster EXC 277 Cognitive Interaction Technology.
* The Excellence Cluster EXC 277 is a grant of the Deutsche
* Forschungsgemeinschaft (DFG) in the context of the German
* Excellence Initiative.
*/

#ifndef ComputeArticulatedScene_H
#define ComputeArticulatedScene_H

#include <stdio.h>

#include <boost/shared_ptr.hpp>
#include <boost/bind.hpp>
#include <boost/thread.hpp>
#include <boost/utility.hpp>

#include <opencv2/core/core.hpp>

#include "ArticulatedScene.h"
#include "XyzImageData.h"

namespace ASM {

/**
 * This class does the actual computations of the Articulated Scene Model
 * algorithm. Use the work() functions in oder to provide data to the
 * algorithm.
 */
class ArticulatedSceneComputer: boost::noncopyable {
public:
	typedef boost::shared_ptr<ArticulatedSceneComputer> Ptr;

	/**
	 * Constructs a new computation object.
	 * @param width The width in pixels of the input images.
	 * @param height The height in pixels of the input images.
	 */
	ArticulatedSceneComputer(int width, int height);
	virtual ~ArticulatedSceneComputer();

	/**
	 * Resets the model. Model will be initialized with a blank background model.
	 */
	virtual void reset();

	/**
	 * Reinitializes the model with a given model.
	 * @param scene The new model for further computations.
	 */
	virtual void initialize(ArticulatedScene::Ptr scene);

	/**
	 * Reinitializes the model with a given background model.
	 * @param backX The x plane of the new background model. Values are
	 * 		expected to be in [mm].
	 * @param backY The y plane of the new background model. Values are
	 * 		expected to be in [mm].
	 * @param backZ The z plane of the new background model. Values are
	 * 		expected to be in [mm].
	 */
	virtual void initialize(cv::Mat backX, cv::Mat backY, cv::Mat backZ);

	/**
	 *  Reinitializes the model with a given background model.
	 * @param background The new background model. Values are
	 * 		expected to be in [mm].
	 */
	virtual void initialize(XyzImageData::Ptr background);

	/**
	 * Do computation based on new input data.
	 * @param input New data to work on. Values are expected to be in [mm].
	 */
	virtual void work(const XyzImageData::Ptr input);

	/**
	 * Do computation based on new input data.
	 * @param xMat values in x dimension [mm].
	 * @param yMat values in y dimension [mm].
	 * @param zMat values in z dimension [mm].
	 */
	virtual void work(const cv::Mat &xMat, const cv::Mat &yMat,
			const cv::Mat &zMat);

	/**
	 * Getter for intermediate results.
	 * @return A copy of the current scene model.
	 */
	virtual ArticulatedScene::Ptr getCurrentSceneCopy() const;

	/**
	 * Getter for intermediate results.
	 * @return A data object containing the static background data.
	 */
	virtual XyzImageData::Ptr getCurrentStatic() const;

	/**
	 * Getter for intermediate results.
	 * @return A data object containing the dynamic foreground data.
	 */
	virtual XyzImageData::Ptr getCurrentDynamic() const;

protected:
	ArticulatedScene::Ptr scene;
	XyzImageData::Ptr currentStatic;
	XyzImageData::Ptr currentDynamic;
	mutable boost::mutex workMutex;
	mutable boost::mutex resultMutex;

	static const int CvMatType = CV_32F;

	int width;
	int height;
};

}

#endif // ComputeArticulatedScene_H
