/*
* This file is part of Articulated Scene Model.
*
* Copyright(c) Agnes Swadzba, Leon Ziegler
* http://opensource.cit-ec.de/projects/asm
*
* This file may be licensed under the terms of the
* GNU Lesser General Public License Version 3 (the ``LGPL''),
* or (at your option) any later version.
*
* Software distributed under the License is distributed
* on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
* express or implied. See the LGPL for the specific language
* governing rights and limitations.
*
* You should have received a copy of the LGPL along with this
* program. If not, go to http://www.gnu.org/licenses/lgpl.html
* or write to the Free Software Foundation, Inc.,
* 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
*
* The development of this software was supported by the
* Excellence Cluster EXC 277 Cognitive Interaction Technology.
* The Excellence Cluster EXC 277 is a grant of the Deutsche
* Forschungsgemeinschaft (DFG) in the context of the German
* Excellence Initiative.
*/

#include "RsbConnectorOut.h"
#include <rsc/misc/langutils.h>

#include <rsb/Factory.h>
#include <rsb/Informer.h>
#include <rst/vision/SimpleXYZImage.pb.h>
#include <iostream>

using namespace rsb;
using namespace rst::vision;
using namespace std;

namespace ASM {

RsbConnectorOut::RsbConnectorOut(const Scope &scope) {
	informer = Factory::getInstance().createInformer<SimpleXYZImage>(scope);
}

RsbConnectorOut::~RsbConnectorOut() {
}

void RsbConnectorOut::put(const XyzImageData::Ptr img) {

	// Explanation: The goal is to send the data without copying the pixel
	// data, but  to ensure that the data is not erased before sending was
	// completed. In order to do this a DataPtr for a new (empty) SimpleXYZImage
	// instance is created (no pixel data is allocated). Afterwards width and
	// height are set and the internal pointers to the data are set to the
	// existing float arrays. To ensure that the incoming shared_ptr does not
	// delete the data before rsb completed the sending process, the new DataPtr
	// is given a custom deleter object. This deleter prevents the original
	// shared_ptr (parent) from deleting anything before the new ptr (child)
	// was destroyed.
	Informer<SimpleXYZImage>::DataPtr dataPtr(new SimpleXYZImage(),
			rsc::misc::ParentSharedPtrDeleter<XyzImageData>(img));
	dataPtr->set_width(img->width());
	dataPtr->set_height(img->height());
	dataPtr->mutable_x()->Reserve(img->width() * img->height());
	dataPtr->mutable_y()->Reserve(img->width() * img->height());
	dataPtr->mutable_z()->Reserve(img->width() * img->height());
	for (int i = 0; i < img->width() * img->height(); i++) {
		dataPtr->mutable_x()->AddAlreadyReserved(img->mutableDataX()[i]);
		dataPtr->mutable_y()->AddAlreadyReserved(img->mutableDataY()[i]);
		dataPtr->mutable_z()->AddAlreadyReserved(img->mutableDataZ()[i]);
	}

	informer->publish(dataPtr);
}

void RsbConnectorOut::put(const RstXyzImageData::Ptr img) {
	Informer<SimpleXYZImage>::DataPtr dataPtr(img->getRstPtr());
	informer->publish(dataPtr);
}

} /* namespace ASM */
