/*
* This file is part of Articulated Scene Model.
*
* Copyright(c) Agnes Swadzba, Leon Ziegler
* http://opensource.cit-ec.de/projects/asm
*
* This file may be licensed under the terms of the
* GNU Lesser General Public License Version 3 (the ``LGPL''),
* or (at your option) any later version.
*
* Software distributed under the License is distributed
* on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
* express or implied. See the LGPL for the specific language
* governing rights and limitations.
*
* You should have received a copy of the LGPL along with this
* program. If not, go to http://www.gnu.org/licenses/lgpl.html
* or write to the Free Software Foundation, Inc.,
* 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
*
* The development of this software was supported by the
* Excellence Cluster EXC 277 Cognitive Interaction Technology.
* The Excellence Cluster EXC 277 is a grant of the Deutsche
* Forschungsgemeinschaft (DFG) in the context of the German
* Excellence Initiative.
*/

#ifndef DetectNearest3DPoint_H
#define DetectNearest3DPoint_H

#include <stdio.h>

#include <boost/shared_ptr.hpp>
#include <boost/bind.hpp>

#include <rsb/Factory.h>
#include <rsb/Scope.h>
#include <rsb/Listener.h>
#include <rsb/Informer.h>
#include <rsb/Event.h>
#include <rsb/Handler.h>

#include <rst/math/Vec3DFloat.pb.h>
#include <rst/vision/SimpleXYZImage.pb.h>

#include <rsb/converter/Repository.h>
#include <rsb/converter/ProtocolBufferConverter.h>

#include <opencv2/highgui/highgui.hpp>
#include <opencv2/core/core.hpp>

class DetectNearest3DPoint {

	typedef boost::shared_ptr<rst::vision::SimpleXYZImage> ImagePtr;

public:

                DetectNearest3DPoint(const rsb::Scope &scope, const rsb::Scope &channel) :
			listener() {
			
		informer = rsb::Factory::getInstance().createInformer<rst::math::Vec3DFloat> (channel);
		
		listener = rsb::Factory::getInstance().createListener(scope);
		listener->addHandler(
				rsb::HandlerPtr(
						new rsb::DataFunctionHandler<rst::vision::SimpleXYZImage>(
								bind(&DetectNearest3DPoint::inputCallback, this,
										_1))));
										
		
	}

	virtual ~DetectNearest3DPoint() {
	}

	virtual void inputCallback(ImagePtr image) {
	
		float *x = (float *) image->mutable_x()->data();
		cv::Mat xMat(image->height(), image->width(), CV_32F, x);
	
		float *y = (float *) image->mutable_y()->data();
		cv::Mat yMat(image->height(), image->width(), CV_32F, y);
	
		float *z = (float *) image->mutable_z()->data();
		cv::Mat zMat(image->height(), image->width(), CV_32F, z);
		
		// compute valid depth measurements (z != 0)
		cv::Mat binMat;
		cv::compare(zMat, 0.0, binMat, cv::CMP_NE);
		
		cv::Point minLoc;
		cv::Point maxLoc;
		double minVal, maxVal;
		
		cv::minMaxLoc(zMat, &minVal, &maxVal, &minLoc, &maxLoc, binMat);
		
		// publish data
		rsb::Informer<rst::math::Vec3DFloat>::DataPtr data(new rst::math::Vec3DFloat());
		data->set_x(xMat.at<float>(minLoc));
		data->set_y(yMat.at<float>(minLoc));
		data->set_z(zMat.at<float>(minLoc));
			
		informer->publish(data);
		
		// build from 1-Channel zMat-image a 3-Channel zMat-image and 
		// mark the nearest 3D point with a red circle
		std::vector<cv::Mat> zVector;
		zVector.push_back(zMat);
		zVector.push_back(zMat);
		zVector.push_back(zMat);
		cv::Mat zMat3(image->height(), image->width(), CV_32FC3);
		cv::merge(zVector, zMat3);
		
		cv::circle(zMat3, minLoc, 3, cv::Scalar(0,0,1));
		
		cv::imshow("minPoint", zMat3);

		return;
		
	}
	
protected:
	rsb::ListenerPtr listener;
	rsb::InformerBasePtr informer;

	
};

#endif // DetectNearest3DPoint_H
