/*
* This file is part of Articulated Scene Model.
*
* Copyright(c) Agnes Swadzba, Leon Ziegler
* http://opensource.cit-ec.de/projects/asm
*
* This file may be licensed under the terms of the
* GNU Lesser General Public License Version 3 (the ``LGPL''),
* or (at your option) any later version.
*
* Software distributed under the License is distributed
* on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
* express or implied. See the LGPL for the specific language
* governing rights and limitations.
*
* You should have received a copy of the LGPL along with this
* program. If not, go to http://www.gnu.org/licenses/lgpl.html
* or write to the Free Software Foundation, Inc.,
* 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
*
* The development of this software was supported by the
* Excellence Cluster EXC 277 Cognitive Interaction Technology.
* The Excellence Cluster EXC 277 is a grant of the Deutsche
* Forschungsgemeinschaft (DFG) in the context of the German
* Excellence Initiative.
*/

#ifndef ExtractConnectedComponents_H
#define ExtractConnectedComponents_H

#include <stdio.h>

#include <boost/shared_ptr.hpp>
#include <boost/bind.hpp>

#include <rsb/Factory.h>
#include <rsb/Scope.h>
#include <rsb/Listener.h>
#include <rsb/Event.h>
#include <rsb/Handler.h>

#include <rst/vision/SimpleXYZImage.pb.h>

#include <rsb/converter/Repository.h>
#include <rsb/converter/ProtocolBufferConverter.h>

#include <opencv2/highgui/highgui.hpp>
#include <opencv2/core/core.hpp>
#include <opencv2/imgproc/imgproc.hpp>

//using namespace cv;

class ExtractConnectedComponents {

	typedef boost::shared_ptr<rst::vision::SimpleXYZImage> ImagePtr;

public:
	ExtractConnectedComponents(const rsb::Scope &scope, const std::string& trackbarname) :
			listener() {
		listener = rsb::Factory::getInstance().createListener(scope);
		listener->addHandler(
				rsb::HandlerPtr(
						new rsb::DataFunctionHandler<rst::vision::SimpleXYZImage>(
								bind(&ExtractConnectedComponents::inputCallback, this,
										_1))));
										
		windowname = trackbarname;

	}

	virtual ~ExtractConnectedComponents() {
	}

	virtual void inputCallback(ImagePtr image) {

		float *a = (float *) image->mutable_z()->data();
		cv::Mat dynamicMat(image->height(), image->width(), CV_32F, a);
		
		// image opening
		cv::Mat openedDynamicMat;
		cv::erode(dynamicMat, openedDynamicMat, cv::Mat::ones(7, 7, CV_32F));
		cv::dilate(openedDynamicMat, openedDynamicMat, cv::Mat::ones(15, 15, CV_32F));
		
		// compute mask of dynamic points
		cv::Mat binMat;
		cv::compare(openedDynamicMat, 0.0, binMat, cv::CMP_NE);
		
		// compute binary edges of the dynamic mask
		cv::Mat bin2Mat;
		cv::dilate(binMat, bin2Mat, cv::Mat::ones(3, 3, CV_32F));
		
		cv::Mat edgeMat;
		cv::subtract(bin2Mat, binMat, edgeMat);
		
		// label not connected dynamic regions
		cv::Mat currBinMat;
		binMat.copyTo(currBinMat);
		cv::Mat currRegionMat;
		
		cv::Mat componentMat = cv::Mat::zeros(image->height(), image->width(), CV_8U);
		
		int counter = 20;
		while(true){
		
			cv::Point seed = findSeed(currBinMat);
			if(seed.x < image->width() && seed.y < image->height()){
				
				edgeMat.copyTo(currRegionMat);
				
				cv::floodFill(currRegionMat, seed, cv::Scalar(255));
				cv::subtract(currRegionMat, edgeMat, currRegionMat);
				
				componentMat.setTo(cv::Scalar(counter), currRegionMat);
				counter += 20;
				
				cv::subtract(currBinMat, currRegionMat, currBinMat);
				
			} else { break; }
		
		}
    		
		cv::imshow(windowname, componentMat);
		
	}
	
protected:
	rsb::ListenerPtr listener;
	std::string windowname;
	
	virtual cv::Point findSeed(cv::Mat binImg){
		
		if(cv::countNonZero(binImg) == 0){
			return cv::Point(binImg.cols, binImg.rows);
		}
		
		int x = 0;
		int y = 0;
		cv::Point p;
		while(true){
			if(binImg.at<uchar>(y,x) == 255){
				p = cv::Point(x,y);
				break;
			} else if(x == binImg.cols-1){
				x = 0;
				y++;
			} else {
				x++;
			}
		}
		return p;
	}
};

#endif // ExtractConnectedComponents_H
