/*
* This file is part of Articulated Scene Model.
*
* Copyright(c) Agnes Swadzba, Leon Ziegler
* http://opensource.cit-ec.de/projects/asm
*
* This file may be licensed under the terms of the
* GNU Lesser General Public License Version 3 (the ``LGPL''),
* or (at your option) any later version.
*
* Software distributed under the License is distributed
* on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
* express or implied. See the LGPL for the specific language
* governing rights and limitations.
*
* You should have received a copy of the LGPL along with this
* program. If not, go to http://www.gnu.org/licenses/lgpl.html
* or write to the Free Software Foundation, Inc.,
* 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
*
* The development of this software was supported by the
* Excellence Cluster EXC 277 Cognitive Interaction Technology.
* The Excellence Cluster EXC 277 is a grant of the Deutsche
* Forschungsgemeinschaft (DFG) in the context of the German
* Excellence Initiative.
*/

#include <asm/ArticulatedSceneComputer.h>

#include "asm_nodes/comm/rsb/RsbServer.h"
#include "asm_nodes/comm/rsb/RsbConnectorIn.h"
#include "asm_nodes/comm/rsb/RsbConnectorOut.h"

#include "asm_nodes/nodes/ComputeArticulatedScene.h"
#include "asm_nodes/nodes/GrabXyzImage.h"
#include "asm_nodes/nodes/VisualizeXyzImage.h"
#include "asm_nodes/nodes/ExtractConnectedComponents.h"
#include "asm_nodes/nodes/DetectNearest3DPoint.h"

#include "tools/info.h"

using namespace rsb;
using namespace rsb::converter;

using namespace rst::vision;
using namespace rst::math;

using namespace boost;
using namespace std;

using namespace ASM;

int main(int argc, char **argv){

	cout << ASM::infoText() << endl;

	// Register a specific template instanciation of the
	// ProtocolBufferConverter for our SimpleXYZImage protocol buffer
	// message.
	shared_ptr<ProtocolBufferConverter<SimpleXYZImage> > converterSimpleXYZImage(
			new ProtocolBufferConverter<SimpleXYZImage>());
	converterRepository<std::string>()->registerConverter(
			converterSimpleXYZImage);

	shared_ptr<ProtocolBufferConverter<Vec3DFloat> > converterVec3DFloat(
			new ProtocolBufferConverter<Vec3DFloat>());
	converterRepository<std::string>()->registerConverter(converterVec3DFloat);

	// Use the RSB factory to create a Server instance that provides
	// callable methods under the scope /bla/server.
	Server::Ptr server(new RsbServer(Scope("/bla/server")));
	ConnectorIn::Ptr connectorImgIn(new RsbConnectorIn(Scope("/bla/raw")));
	ConnectorOut::Ptr connectorImgOut(new RsbConnectorOut(Scope("/bla/raw")));
	ConnectorIn::Ptr connectorStaticIn(
			new RsbConnectorIn(Scope("/bla/static")));
	ConnectorOut::Ptr connectorStaticOut(
			new RsbConnectorOut(Scope("/bla/static")));
	ConnectorIn::Ptr connectorDynamicIn(
			new RsbConnectorIn(Scope("/bla/dynamic")));
	ConnectorOut::Ptr connectorDynamicOut(
			new RsbConnectorOut(Scope("/bla/dynamic")));

	// Initialize the Articulated Scene Modeling
	ArticulatedSceneComputer asmComp(640, 480);

	// Create Nodes
	ComputeArticulatedScene asmNode(asmComp);
	asmNode.initCommunication(connectorImgIn, connectorStaticOut,
			connectorDynamicOut, server);

	// Create grabbing node
	GrabXyzImage grabNode(connectorImgOut, 640, 480, 30);

	// Visualize raw depth data and extracted static and dynamic data
	cv::namedWindow("depth", CV_WINDOW_AUTOSIZE);
	VisualizeXyzImage vizDepthNode(connectorImgIn, "depth");

	cv::namedWindow("static", CV_WINDOW_AUTOSIZE);
	VisualizeXyzImage vizStaticNode(connectorImgIn, "static");

	cv::namedWindow("dynamic", CV_WINDOW_AUTOSIZE);
	VisualizeXyzImage vizDynamicNode(connectorImgIn, "dynamic");

	// Segment 3D points marked as dynamic using connected components in 2D
	cv::namedWindow("components", CV_WINDOW_AUTOSIZE);
	boost::shared_ptr<ExtractConnectedComponents> compimg(
			new ExtractConnectedComponents(Scope("/bla/dynamic"),
					"components"));

	// Detect nearest Point in raw image.
	cv::namedWindow("minPoint", CV_WINDOW_AUTOSIZE);
	boost::shared_ptr<DetectNearest3DPoint> nearimg(
			new DetectNearest3DPoint(Scope("/bla/raw"), Scope("/bla/nearest")));

	cout << " done" << endl;
	cout << "press any key to quit" << endl;

	cv::waitKey();

	return 0;
}

